/*
 *  Path.es --  Path class. Path's represent files in a file system. The file may or may not exist.  
 *
 *  Copyright (c) All Rights Reserved. See details at the end of the file.
 */

/*
 *  FUTURE
 *      Path.expand(hash: Object = null)        Expand $VARS or ${vars} in path. Also expand ~user
 *      Path.split(): Array                     Return dir + base. Useful if you have destructuring assignment
 *      Path.splitDrive(): Array                Return drive + rest
 *      Path.splitExtension(): Array            Return rest + extension
 *      Path.splitUNC(): Array                  Return unc + rest (\\host)
 *      Path.normalizeCase                      Map to lower case
 *      Path.canonicalize()                     Resolve symbolic links and return normalized, validated, fully resolved path.
 *                                              Error if the path does not exist.
 *      Path.backup(): Path                     do a back and return the name of the backup.
 *      Path.crc(): Number                      Calculate a 32 bit crc of the file contents
 *      Path.md5(): Number                      Calculate an MD5 digest of the file contents
 *      Path.touch(date: Date = null): Void     Update modified time. Create if non-existant
 *      Path.perms = NNN                        Permissions getter / setter
 *      Path.append(line)                       Append a line to a file (open, write, close)
 *      Path.isReadable, isWritable, isHidden, isSpecial, is Mount
 *      Path.files(/regexp/)                    
 *      Path.rmtree()                           Recursive removal
 *      Path.validate                           Validate a path.
 */

module ejs.io {

    /**
     *  Path class. Paths represent files in a file system. The corresponding file or directory for the Path  may or 
     *  may not exist. The representation of a path can be either natural, where the path delimiter is the normal 
     *  O/S path delimiter. For example, on Windows the path delimiter is "\" and on Unix it is "/". Paths can also
     *  be viewed with a unified representation where the delimiter is always "/". Ejscript typically operates in
     *  mapped representation to assist creating portable applications. The representaion can be changed via the
     *  representation() method.
     *  @spec ejs-12
     */
    native class Path {

        use default namespace public

        static const NATIVE = 1
        static const PORTABLE  = 2

        /**
         *  Create a new Path object and set the path's location.
         *  @param path the name of the file to associate with this path object. 
         */
        native function Path(path: String)

        /**
         *  Return an absolute path name.
         *  @return an absolute path equivalent for the current path. The returned path is normalized.
         */
        native function get absolute(): Path

        /**
         *  Get when the file was last accessed.
         *  @return A date object with the date and time or null if the method fails.
         */
        native function get accessed(): Date 

        /**
         *  Get the file security permissions.
         *  @return the file attributes object hash. Fields include: 
         *  @options owner String representing the file owner
         *  @options group String representing the file group
         *  @options permissions Number File Posix permissions mask
         */
        #FUTURE
        native function get attributes(): Object

        /**
         *  Set the file security permissions. 
         *  @params attributes Attribute object hash. Options include:
         *  @options owner String representing the file owner
         *  @options group String representing the file group
         *  @options permissions Number File Posix permissions mask
         */
        #FUTURE
        native function set attributes(attributes: Object): Void

        /**
         *  Get the base of portion of the path. The base portion is the trailing portion without any directory elements.
         *  @return A path containing the base name portion of the current path.
         */
        native function get basename(): Path
        
        /**
         *  Convert a path to an absolute path and break into components
         *  @return an array with an element for each segment of the path. The first element represents the
         *      the file system root and will be the empty string or drive spec on a windows like system.
         */
        function get components(): Array
            absolute.toString().split(separator)

        /**
         *  Copy a file
         *  @param target New file location
         *  @param options Object hash
         *  @options permissions Set to a numeric Posix permissions mask
         */
        native function copy(target: Object, options: Object = null): Void

        /**
         *  Get when the file was created.
         *  @return A date object with the date and time or null if the file does not exist.
         *  @rename accessed modified 
         */
        native function get created(): Date 

        /**
         *  Get the directory portion of a file. The directory portion is the leading portion including all 
         *  directory elements and excluding the base name. On some systems, it will include a drive specifier.
         *  @return A string containing the directory name portion of the path.
         */
        native function get dirname(): Path

        /**
         *  Return true if the path ends with the given suffix
         *  @param suffix Path or String suffix to compare with the path.
         *  @return true if the path does begin with the suffix
         */
        function endsWith(suffix: Object): Boolean
            this.toString().endsWith(suffix.toString())

        /**
         *  Test to see if a file for this path exists.
         *  @return True if the file exists.
         */
        native function get exists(): Boolean 

        /**
         *  Get the file extension portion of the path.
         *  @return String containing the file extension
         */
        native function get extension(): String 

        /**
         *  Get a list of files in a directory. The returned array contains the base path portion only.
         *  @param path Directory path to enumerate.
         *  @param enumDirs If set to true, then dirList will include sub-directories in the returned list of files.
         *  @return An Array of strings containing the filenames in the directory.
         */
        native function files(enumDirs: Boolean = false): Array 

        /**
         *  Get the file system containing this Path 
         *  @returns The file system object containing this path.
         */
        #FUTURE
        native function get fileSystem(): FileSystem

        /**
         *  Get iterate over any files contained under this path (assuming it is a directory) "for (v in files)". 
         *      This operates the same as getValues on a Path object.
         *  @param deep Follow the prototype chain
         *  @return An iterator object.
         *  @example:
         *      for (f in Path("."))
         */
        override iterator native function get(deep: Boolean = false): Iterator

        /**
         *  Get an iterator for this file to be used by "for each (v in obj)". Return 
         *      This operates the same as $get on a Path object.
         *  @param deep Follow the prototype chain. Only implemented in ECMA compliance mode..
         *  @return An iterator object.
         *  @example
         *      for each (f in Path("."))
         */
        override iterator native function getValues(deep: Boolean = false): Iterator

        /**
         *  Determine if the file path has a drive spec (C:) in it's name. Only relevant on Windows like systems.
         *  @return True if the file path has a drive spec
         */
        native function get hasDrive(): Boolean 

        /**
         *  Determine if the path is absolute
         *  @return true if the path is an absolute path beginning at the file system's root.
         */
        native function get isAbsolute(): Boolean

        /**
         *  Determine if the path is a directory
         *  @return true if the file is a directory
         */
        native function get isDir(): Boolean

        /**
         *  Determine if the path is a link file (either symbolic or hard)
         *  @return true if the file is a link
         */
        native function get isLink(): Boolean

        /**
         *  Determine if the path is a regular file
         *  @return true if the file is a regular file and not a directory
         */
        native function get isRegular(): Boolean

        /**
         *  Determine if the path is absolute
         *  @return true if the path is an absolute path beginning at the file system's root.
         */
        native function get isRelative(): Boolean

        /**
         *  Determine if the path is a symbolic link file
         *  @return true if the file is a symbolic link
         */
        native function get isSymbolic(): Boolean

        /**
         *  Join paths. Joins one more more paths together. If another path is absolute, return it. Otherwise
         *  add a separator, and continue joining.
         *  @return A joined, normalized path.
         */
        native function get join(...other): Path

        /**
         *  Join an extension to a path. If the base name of the path already has an extension, this call does nothing.
         *  @return A path with extension.
         */
        native function get joinExt(ext: String): Path

        /**
         *  Get the length of the path name. 
         *  @return The number of characters in the path name.
         */
        override native function get length(): Number 

        /**
         *  Make a new directory and all required intervening directories. If the directory already exists, 
         *      the function returns without throwing an exception.
         *  @param path Filename path to use.
         *  @param options
         *  @options permissions Set to a numeric Posix permissions mask
         *  @options owner String representing the file owner (Not implemented)
         *  @options group String representing the file group (Not implemented)
         *  @throws IOError if the directory cannot be created.
         */
        native function makeDir(options: Object = null): void

        /**
         *  Create a link to a file
         *  @param existing Path to an existing file to link to.
         *  @param hard Set to true to create a hard link. Otherwise the default is to create a symbolic link.
         *  @returns this path
         */
        native function makeLink(existing: Path, hard: Boolean = false): Path

        /**
         *  Create a temporary file in the path directory. Creates a new, uniquely named temporary file.
         *  @returns a new Path object for the temp file.
         */
        native function makeTemp(): Path

        /**
         *  Get when the file was created or last modified.
         *  @return A date object with the date and time or null if the method fails.
         */
        native function get modified(): Date 

        /**
         *  Get the name of the Path as a string in the default representation. This is the same as $toString.
         *  @return The name of the path.
         */
        native function get name(): String 

        /**
         *  Get the path in the O/S natural (native) representation. This uses the native O/S separators, this is
         *      "\" on windows and "/" on unix.
         *  @return the path in the O/S natural representation. The returned path is normalized.
         */
        native function get natural(): Path 

        /**
         *  Normalize a path by removing all redundant and invalid path components. This removes "/..", "segment/.." 
         *  and "./" components. It not convert to an absolute path nor will it map the case of the filename.
         *  @return a normalized path with all 
         */
        native function get normalize(): Path

        /**
         *  Open a path and return a File object.
         *  @params options
         *  @options mode optional file access mode string. Use "r" for read, "w" for write, "a" for append to existing
         *      content, "+" never truncate.
         *  @options permissions optional Posix permissions number mask. Defaults to 0664.
         *  @options owner String representing the file owner (Not implemented)
         *  @options group String representing the file group (Not implemented)
         *  @return a File object which implements the Stream interface.
         *  @throws IOError if the path or file cannot be created.
         */
        function open(options: Object = null): File
            new File(this, options)

        /**
         *  Open a file and return a TextStream object.
         *  @params options Optional options object
         *  @options mode optional file access mode string. Use "r" for read, "w" for write, "a" for append to existing
         *      content, "+" never truncate.
         *  @options encoding Text encoding format
         *  @options permissions optional Posix permissions number mask. Defaults to 0664.
         *  @options owner String representing the file owner (Not implemented)
         *  @options group String representing the file group (Not implemented)
         *  @return a TextStream object which implements the Stream interface.
         *  @throws IOError if the path or file cannot be opened or created.
         */
        function openTextStream(options: Object = null): TextStream
            new TextStream(open(options))

        /**
         *  Open a file and return a BinaryStream object.
         *  @params options Optional options object
         *  @param mode optional file access mode with values ored from: Read, Write, Append, Create, Open, Truncate. 
         *      Defaults to Read.
         *  @options permissions optional Posix permissions number mask. Defaults to 0664.
         *  @options owner String representing the file owner (Not implemented)
         *  @options group String representing the file group (Not implemented)
         *  @return a BinaryStream object which implements the Stream interface.
         *  @throws IOError if the path or file cannot be opened or created.
         */
        function openBinaryStream(options: Object = null): BinaryStream
            new BinaryStream(open(options))

        /**
         *  Get the parent directory of path. This may need to convert to an absolute path if there are no parent
         *  directories in a relative path.
         *  @return the parent directory as a Path
         */
        native function get parent(): Path

        /**
         *  Get the path in the portable (like Unix) representation. This uses "/" separators.
         *  @param lower If set to true, all characters in the path will be mapped to lower case.
         *  @return the path in a portable representation. The returned path is normalized.
         */
        native function get portable(lower: Boolean = false): Path 

        /**
         *  Read the file contents as an array of lines. Each line is a string. This method opens the file, 
         *      reads the contents and closes the file.
         *  @param path the name of the file to read.
         *  @return An array of strings.
         *  @throws IOError if the file cannot be read
         *  @example:
         *      for each (byte in Path("/tmp/a.txt").readBytes())
         */
        function readBytes(): ByteArray {
            var file: File = File(this).open()
            result = file.readBytes()
            file.close()
            return result
        }

        /**
         *  Read the file contents as an array of lines. Each line is a string. This method opens the file, reads the 
         *      contents and closes the file.
         *  @param path the name of the file to read.
         *  @return An array of strings.
         *  @throws IOError if the file cannot be read
         *  @example:
         *      for each (line in Path("/tmp/a.txt").readLines())
         */
        function readLines(): Array {
            var stream: TextStream = TextStream(open(this))
            result = stream.readLines()
            stream.close()
            return result
        }

        /**
         *  Read the file contents as a string. This method opens the file, reads the contents and closes the file.
         *  @param path the name of the file to read.
         *  @return A string.
         *  @throws IOError if the file cannot be read
         *  @example:
         *      Path("/tmp/a.txt").readString()
         */
        function readString(): String {
            var file: File = open(this)
            result = file.readString()
            file.close()
            return result
        }

        /**
         *  Read the file contents as an XML object.  This method opens the file, reads the contents and closes the file.
         *  @param path the name of the file to read.
         *  @return An XML object
         *  @throws IOError if the file cannot be read
         */
        function readXML(): XML {
            var file: File = open(this)
            data = file.readString()
            if (data == null) {
                return null
            }
            file.close()
            return new XML(data)
        }

        /**
         *  Return a relative path name for the file.
         *  @return a string containing a file path name relative to the application's current working directory..
         *  The returned path is normalized.
         */
        native function get relative(): Path

        /**
         *  Delete the file associated with the File object. 
         *  @throws IOError if the file exists and could not be deleted.
         */
        native function remove(): void 

        /**
         *  Removes a directory.
         *  @param path Filename path to remove.
         *  @param recursive If true, remove all subdirectories with their contents
         *  @throws IOError if the directory exists and cannot be removed.
         */
        native function removeDir(recursive: Boolean = false): void

        /**
         *  Rename a file. If the new path exists it is removed before the rename.
         *  @param target New name of the path. Can be either a Path or String.
         *  @throws IOError if the original file does not exist or cannot be renamed.
         */
        native function rename(target: Object): void
        
        /**
         *  Get the path name representation format
         *  @returns the Path format. Will return either: NATIVE or PORTABLE.
         */
        native static function get representation(): Number

        /**
         *  Set the path name representation format
         *  @param format Path format. Must be either: NATIVE or PORTABLE.
         */
        native static function set representation(format: Number): Void

        /**
         *  Compare two paths test if they represent the same file
         *  @param other Other path to compare with
         *  @return True if the paths represent the same underlying filename
         */
        native function same(other: Object): Boolean

        /**
         *  Return the path separator for this path
         *  @return A string. Will typically be either "/" or "\\" depending on the platform and representation.
         */
        native function get separator(): String

        /**
         *  Get the size of the file associated with this Path object.
         *  @return The number of bytes in the file or -1 if length determination failed.
         */
        native function get size(): Number 

        /**
         *  Return true if the path starts with the given prefix
         *  @param prefix Path or String prefix to compare with the path.
         *  @return true if the path does begin with the prefix
         */
        function startsWith(prefix: Object): Boolean
            this.toString().startsWith(prefix.toString()) 

        /**
         *  Convert the path to a JSON string. 
         *  @return a JSON string representing the path.
         */
        native override function toJSON(): String

        /**
         *  Convert the path to a string. The format of the string will depend on the defined $representation format.
         *  @return a string representing the path.
         */
        native override function toString(): String

        //  TODO - should support reg expressions
        /**
         *  Trim a pattern from the end of the path
         *  NOTE: this does a case-sensitive match
         *  @return a Path containing the trimmed path name
         */
        function trimEnd(pat: String): Path {
            if (name.endsWith(pat)) {
                loc = name.lastIndexOf(pat)
                if (loc >= 0) {
                    return new Path(name.slice(0, loc))
                }
            }
            return this
        }

        /**
         *  Trim a file extension from a path
         *  @return a Path with no extension
         */
        native function trimExt(): Path

        //  TODO - should support reg expressions
        /**
         *  Trim a pattern from the start of the path
         *  NOTE: this does a case-sensitive match
         *  @return a Path containing the trimmed path name
         */
        function trimStart(pat: String): Path {
            if (name.startsWith(pat)) {
                return new Path(name.slice(pat.length))
            }
            return this
        }

        /**
         *  Reduce the size of the file by truncating it. 
         *  @param size The new size of the file. If the truncate argument is greater than or equal to the 
         *      current file size nothing happens.
         *  @throws IOError if the truncate failed.
         */
        native function truncate(size: Number): void

        /**
         *  Write the file contents. This method opens the file, writes the contents and closes the file.
         *  @param args The data to write to the file. Data is serialized in before writing. Note that numbers will not 
         *      be written in a cross platform manner. If that is required, use the BinaryStream class to write Numbers.
         *  @throws IOError if the data cannot be written
         */
        function write(...args): void {
            var file: File

            try {
                file = new File(this, { mode: "w", permissions: 0644 })
                for each (item in args) {
                    if (item is String) {
                        file.write(item)
                    } else {
                        file.write(serialize(item))
                    }
                }
            } finally {
                file.close()
            }
        }
    }
}

/*
 *  @copy   default
 *  
 *  Copyright (c) Embedthis Software LLC, 2003-2009. All Rights Reserved.
 *  Copyright (c) Michael O'Brien, 1993-2009. All Rights Reserved.
 *  
 *  This software is distributed under commercial and open source licenses.
 *  You may use the GPL open source license described below or you may acquire 
 *  a commercial license from Embedthis Software. You agree to be fully bound 
 *  by the terms of either license. Consult the LICENSE.TXT distributed with 
 *  this software for full details.
 *  
 *  This software is open source; you can redistribute it and/or modify it 
 *  under the terms of the GNU General Public License as published by the 
 *  Free Software Foundation; either version 2 of the License, or (at your 
 *  option) any later version. See the GNU General Public License for more 
 *  details at: http://www.embedthis.com/downloads/gplLicense.html
 *  
 *  This program is distributed WITHOUT ANY WARRANTY; without even the 
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  
 *  This GPL license does NOT permit incorporating this software into 
 *  proprietary programs. If you are unable to comply with the GPL, you must
 *  acquire a commercial license to use this software. Commercial licenses 
 *  for this software and support services are available from Embedthis 
 *  Software at http://www.embedthis.com 
 *  
 *  @end
 */
