/**
 *  ejsPath.c - Path class.
 *
 *  Copyright (c) All Rights Reserved. See details at the end of the file.
 */

/********************************** Includes **********************************/

#include    "ejs.h"

#if ES_ejs_io_Path
/************************************ Forwards ********************************/

static char *getPath(Ejs *ejs, EjsVar *vp);

/************************************ Methods *********************************/
/*
 *  Constructor
 *
 *  function Path(path: String)
 *
 */
static EjsVar *pathConstructor(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    mprAssert(argc == 1);
#if UNUSED
    fp->path = mprGetNormalizedPath(fp, path);
#endif
    if ((fp->path = getPath(ejs, argv[0])) == 0) {
        return (EjsVar*) fp;
    }
    fp->path = mprStrdup(fp, fp->path);
    return (EjsVar*) fp;
}


/*
 *  Return an absolute path name for the file
 *
 *  function get absolutePath()
 */
static EjsVar *absolutePath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetAbsPath(fp, fp->path));
}


/*
 *  Get when the file was last accessed.
 *
 *  function get accessed(): Date
 */
static EjsVar *getAccessedDate(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;

    mprGetPathInfo(ejs, fp->path, &info);
    return (EjsVar*) ejsCreateDate(ejs, (MprTime) info.atime);
}


/*
 *  Get the base name of a file
 *
 *  function basename(): Path
 */
static EjsVar *getPathBasename(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetPathBase(ejs, fp->path));
}


/*
 *  Copy a file
 *
 *  function copy(to: Object, options: Object = null): Void
 *
 *  TODO - not implementing copy options parameter.
 */
static EjsVar *copyPath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprFile     *from, *to;
    char        *buf, *toPath;
    uint        bytes;
    int         rc;

    mprAssert(argc == 1);
    if ((toPath = getPath(ejs, argv[0])) == 0) {
        return 0;
    }

    from = mprOpen(ejs, fp->path, O_RDONLY | O_BINARY, 0);
    if (from == 0) {
        ejsThrowIOError(ejs, "Cant open %s", fp->path);
        return 0;
    }

    to = mprOpen(ejs, toPath, O_CREAT | O_WRONLY | O_TRUNC | O_BINARY, EJS_FILE_PERMS);
    if (to == 0) {
        ejsThrowIOError(ejs, "Cant create %s", toPath);
        mprFree(from);
        return 0;
    }

    buf = mprAlloc(ejs, MPR_BUFSIZE);
    if (buf == 0) {
        ejsThrowMemoryError(ejs);
        mprFree(to);
        mprFree(from);
        return 0;
    }

    rc = 0;
    while ((bytes = mprRead(from, buf, MPR_BUFSIZE)) > 0) {
        if (mprWrite(to, buf, bytes) != bytes) {
            ejsThrowIOError(ejs, "Write error to %s", toPath);
            rc = 0;
            break;
        }
    }
    mprFree(from);
    mprFree(to);
    mprFree(buf);
    return 0;
}


/*
 *  Return when the file was created.
 *
 *  function get created(): Date
 */
static EjsVar *getCreatedDate(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;

    mprGetPathInfo(ejs, fp->path, &info);
    return (EjsVar*) ejsCreateDate(ejs, (MprTime) info.ctime);
}


/**
 *  Get the directory name portion of a file.
 *
 *  function get dirname(): Path
 */
static EjsVar *getPathDirname(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetPathDir(ejs, fp->path));
}


/*
 *  Test to see if this file exists.
 *
 *  function get exists(): Boolean
 */
static EjsVar *getPathExists(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;

    return (EjsVar*) ejsCreateBoolean(ejs, mprGetPathInfo(ejs, fp->path, &info) == 0);
}


/*
 *  Get the file extension portion of the file name.
 *
 *  function get extension(): String
 */
static EjsVar *getPathExtension(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    char    *cp;

    if ((cp = strrchr(fp->path, '.')) == 0) {
        return (EjsVar*) ejs->emptyStringValue;
    }
    return (EjsVar*) ejsCreateString(ejs, cp);
}


/*
 *  Function to iterate and return the next element index.
 *  NOTE: this is not a method of Array. Rather, it is a callback function for Iterator
 */
static EjsVar *nextPathKey(Ejs *ejs, EjsIterator *ip, int argc, EjsVar **argv)
{
    EjsPath     *fp;

    fp = (EjsPath*) ip->target;
    if (!ejsIsFile(fp)) {
        ejsThrowReferenceError(ejs, "Wrong type");
        return 0;
    }

    if (ip->index < mprGetListCount(fp->files)) {
        return (EjsVar*) ejsCreateNumber(ejs, ip->index++);
    }
    ejsThrowStopIteration(ejs);
    return 0;
}


/*
 *  Return the default iterator for use with "for ... in". This will iterate over the files in a directory.
 *
 *  iterator native function get(): Iterator
 */
static EjsVar *getPathIterator(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    mprFree(fp->files);
    fp->files = mprGetPathFiles(fp, fp->path, 0);
    return (EjsVar*) ejsCreateIterator(ejs, (EjsVar*) fp, (EjsNativeFunction) nextPathKey, 0, NULL);
}


/*
 *  Function to iterate and return the next element value.
 *  NOTE: this is not a method of Array. Rather, it is a callback function for Iterator
 */
static EjsVar *nextPathValue(Ejs *ejs, EjsIterator *ip, int argc, EjsVar **argv)
{
    EjsPath     *fp;

    fp = (EjsPath*) ip->target;
    if (!ejsIsFile(fp)) {
        ejsThrowReferenceError(ejs, "Wrong type");
        return 0;
    }

    if (ip->index < mprGetListCount(fp->files)) {
        return (EjsVar*) ejsCreatePath(ejs, mprGetItem(fp->files, ip->index++));
    }
    ejsThrowStopIteration(ejs);
    return 0;
}


/*
 *  Return an iterator to enumerate the bytes in the file. For use with "for each ..."
 *
 *  iterator native function getValues(): Iterator
 */
static EjsVar *getPathValues(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    mprFree(fp->files);
    fp->files = mprGetPathFiles(fp, fp->path, 0);
    return (EjsVar*) ejsCreateIterator(ejs, (EjsVar*) fp, (EjsNativeFunction) nextPathValue, 0, NULL);
}


/*
 *  Get the files in a directory.
 *  function getFiles(enumDirs: Boolean = false): Array
 *
 *  TODO -- need pattern to match (what about "." and ".." and ".*")
 *  TODO - move this functionality into mprFile (see appweb dirHandler.c)
 */
static EjsVar *getPathFiles(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    char            path[MPR_MAX_FNAME];
    EjsArray        *array;
    MprList         *list;
    MprDirEntry     *dp;
    bool            enumDirs, noPath;
    int             next;

    mprAssert(argc == 0 || argc == 1);

    enumDirs = (argc == 1) ? ejsGetBoolean(argv[0]): 0;

    array = ejsCreateArray(ejs, 0);
    if (array == 0) {
        //  TODO - great to push this down into ejsAllocVar
        ejsThrowMemoryError(ejs);
        return 0;
    }

    list = mprGetPathFiles(array, fp->path, enumDirs);
    if (list == 0) {
        ejsThrowIOError(ejs, "Can't read directory");
        return 0;
    }

    noPath = (fp->path[0] == '.' && fp->path[1] == '\0') || 
        (fp->path[0] == '.' && fp->path[1] == '/' && fp->path[2] == '\0');

    for (next = 0; (dp = mprGetNextItem(list, &next)) != 0; ) {
        if (strcmp(dp->name, ".") == 0 || strcmp(dp->name, "..") == 0) {
            continue;
        }
        if (enumDirs || !(dp->isDir)) {
            if (noPath) {
                ejsSetProperty(ejs, (EjsVar*) array, -1, (EjsVar*) ejsCreatePath(ejs, dp->name));
            } else {
                /*
                 *  Prepend the directory name
                 */
                mprSprintf(path, sizeof(path), "%s/%s", fp->path,  dp->name);
                ejsSetProperty(ejs, (EjsVar*) array, -1, (EjsVar*) ejsCreatePath(ejs, path));
            }
        }
    }
    mprFree(list);

    return (EjsVar*) array;
}


#if FUTURE
static EjsVar *fileSystem(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    //  TODO
    return 0;
}
#endif


/*
 *  Determine if the file path has a drive spec (C:) in the file name
 *
 *  static function hasDrive(): Boolean
 */
static EjsVar *pathHasDrive(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreateBoolean(ejs, 
        (isalpha((int) fp->path[0]) && fp->path[1] == ':' && (fp->path[2] == '/' || fp->path[2] == '\\')));
}


/*
 *  function get isAbsolute(): Boolean
 */
static EjsVar *isPathAbsolute(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) (mprIsAbsPath(ejs, fp->path) ? ejs->trueValue: ejs->falseValue);
}


/*
 *  Determine if the file name is a directory
 *
 *  function get isDir(): Boolean
 */
static EjsVar *isPathDir(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;
    int         rc;

    rc = mprGetPathInfo(ejs, fp->path, &info);

    return (EjsVar*) ejsCreateBoolean(ejs, rc == 0 && info.isDir);
}


/*
 *  function get isLink(): Boolean
 */
static EjsVar *isPathLink(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    //  TODO
    mprAssert(0);
    return 0;
}


/*
 *  Determine if the file name is a regular file
 *
 *  function get isRegular(): Boolean
 */
static EjsVar *isPathRegular(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;

    mprGetPathInfo(ejs, fp->path, &info);

    return (EjsVar*) ejsCreateBoolean(ejs, info.isReg);
}


/*
 *  function get isRelative(): Boolean
 */
static EjsVar *isPathRelative(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) (mprIsRelPath(ejs, fp->path) ? ejs->trueValue: ejs->falseValue);
}


/*
 *  function get isSymbolic(): Boolean
 */
static EjsVar *isPathSymbolic(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    //  TODO
    mprAssert(0);
    return 0;
}


/*
 *  Join paths. Returns a normalized path.
 *
 *  function join(...others): Path
 */
static EjsVar *joinPath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    EjsArray    *args;
    cchar       *other, *join;
    int         i;

    args = (EjsArray*) argv[0];
    join = fp->path;
    for (i = 0; i < args->length; i++) {
        if ((other = getPath(ejs, ejsGetProperty(ejs, (EjsVar*) args, i))) == NULL) {
            return 0;
        }
        join = mprJoinPath(ejs, join, other);
    }
    return (EjsVar*) ejsCreatePath(ejs, join);;
}


/*
 *  Join extension
 *
 *  function joinExt(ext: String): Path
 */
static EjsVar *joinPathExt(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    if (mprGetPathExtension(ejs, fp->path)) {
        return (EjsVar*) fp;
    }
    return (EjsVar*) ejsCreatePath(ejs, mprStrcat(ejs, -1, fp->path, ejsGetString(argv[0]), NULL));
}


/*
 *  Get the length of the path name.
 *
 *  override intrinsic function get length(): Number
 */
static EjsVar *pathLength(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreateNumber(ejs, strlen(fp->path));
}


/*
 *  function makeDir(options: Object = null): Void
 *
 *  Options: permissions, owner, group
 */
static EjsVar *makePathDir(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;
    EjsName     qname;
    EjsVar      *options, *permissions;
#if FUTURE
    EjsVar      *owner, *group;
    cchar       *ownerName, *groupName;
#endif
    int         perms;
    
    perms = 0755;

    if (argc == 1) {
        if (!ejsIsObject(argv[0])) {
            ejsThrowArgError(ejs, "Bad args");
        }
        options = argv[0];

        if (ejsIsObject(options)) {
            permissions = ejsGetPropertyByName(ejs, options, ejsName(&qname, EJS_PUBLIC_NAMESPACE, "permissions"));
    #if FUTURE
            owner = ejsGetPropertyByName(ejs, options, ejsName(&qname, EJS_PUBLIC_NAMESPACE, "owner"));
            group = ejsGetPropertyByName(ejs, options, ejsName(&qname, EJS_PUBLIC_NAMESPACE, "group"));
    #endif
            if (permissions) {
                perms = ejsGetInt(permissions);
            }
        }
    }
    if (mprGetPathInfo(ejs, fp->path, &info) == 0 && info.isDir) {
        return 0;
    }
    if (mprMakeDir(ejs, fp->path, perms, 1) < 0) {
        ejsThrowIOError(ejs, "Cant create directory %s", fp->path);
        return 0;
    }
#if FUTURE
    if (owner) {
        ownerName = ejsGetString(owner);
    }
    if (group) {
        groupName = ejsGetString(group);
    }
#endif
    return 0;
}


/*
 *  function makeLink(existing: Path, hard: Boolean = false): Path
 */
static EjsVar *makePathLink(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    //  TODO 
    mprAssert(0);
    return 0;
}


/*
 *  Make a temporary file. Creates a new, uniquely named temporary file.
 *
 *  function makeTemp(): Path
 */
static EjsVar *makePathTemp(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    char    *path;

    if ((path = mprGetTempPath(ejs, fp->path)) == NULL) {
        ejsThrowIOError(ejs, "Can't make temp file");
        return 0;
    }
    return (EjsVar*) ejsCreatePathAndFree(ejs, path);
}


/*
 *  Get when the file was created or last modified.
 *
 *  function get modified(): Date
 */
static EjsVar *getModifiedDate(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;

    mprGetPathInfo(ejs, fp->path, &info);
    return (EjsVar*) ejsCreateDate(ejs, (MprTime) info.mtime);
}


/*
 *  Get the name of the path as a string.
 *
 *  function get name(): String
 */
static EjsVar *getPathName(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreateString(ejs, fp->path);
}


/*
 *  function get natural(): Path
 */
static EjsVar *getNaturalPath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetNativePath(ejs, fp->path));
}


/*
 *  function get normalize(): Path
 */
static EjsVar *normalizePath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetNormalizedPath(ejs, fp->path));
}


/*
 *  Get the parent directory of the absolute path of the file.
 *
 *  function get parent(): String
 */
static EjsVar *getPathParent(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetPathParent(fp, fp->path));
}


/*
 *  Get a portable (unix-like) representation of the path
 *
 *  function get portable(): Path
 */
static EjsVar *getPortablePath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetPortablePath(ejs, fp->path));
}


#if UNUSED
/*
 *  Get the file contents as a byte array
 *
 *  static function readBytes(path: String): ByteArray
 */
static EjsVar *readBytes(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprFile         *file;
    EjsByteArray    *result;
    cchar           *path;
    char            buffer[MPR_BUFSIZE];
    int             bytes, offset, rc;

    mprAssert(argc == 1 && ejsIsString(argv[0]));
    path = ejsGetString(argv[0]);

    file = mprOpen(ejs, path, O_RDONLY | O_BINARY, 0);
    if (file == 0) {
        ejsThrowIOError(ejs, "Can't open %s", path);
        return 0;
    }

    /*
     *  TODO - need to be smarter about running out of memory here if the file is very large.
     */
    result = ejsCreateByteArray(ejs, (int) mprGetFileSize(file));
    if (result == 0) {
        ejsThrowMemoryError(ejs);
        return 0;
    }

    rc = 0;
    offset = 0;
    while ((bytes = mprRead(file, buffer, MPR_BUFSIZE)) > 0) {
        if (ejsCopyToByteArray(ejs, result, offset, buffer, bytes) < 0) {
            ejsThrowMemoryError(ejs);
            rc = -1;
            break;
        }
        offset += bytes;
    }
    ejsSetByteArrayPositions(ejs, result, 0, offset);

    mprFree(file);
    return (EjsVar*) result;
}


/**
 *  Read the file contents as an array of lines.
 *
 *  static function readLines(path: String): Array
 */
static EjsVar *readLines(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprFile     *file;
    MprBuf      *data;
    EjsArray    *result;
    cchar       *path;
    char        *start, *end, *cp, buffer[MPR_BUFSIZE];
    int         bytes, rc, lineno;

    mprAssert(argc == 1 && ejsIsString(argv[0]));
    path = ejsGetString(argv[0]);

    result = ejsCreateArray(ejs, 0);
    if (result == NULL) {
        ejsThrowMemoryError(ejs);
        return 0;
    }

    file = mprOpen(ejs, path, O_RDONLY | O_BINARY, 0);
    if (file == 0) {
        ejsThrowIOError(ejs, "Can't open %s", path);
        return 0;
    }

    /*
     *  TODO - need to be smarter about running out of memory here if the file is very large.
     */
    data = mprCreateBuf(ejs, 0, (int) mprGetFileSize(file) + 1);
    result = ejsCreateArray(ejs, 0);
    if (result == NULL || data == NULL) {
        ejsThrowMemoryError(ejs);
        mprFree(file);
        return 0;
    }

    rc = 0;
    while ((bytes = mprRead(file, buffer, MPR_BUFSIZE)) > 0) {
        if (mprPutBlockToBuf(data, buffer, bytes) != bytes) {
            ejsThrowMemoryError(ejs);
            rc = -1;
            break;
        }
    }

    start = mprGetBufStart(data);
    end = mprGetBufEnd(data);
    for (lineno = 0, cp = start; cp < end; cp++) {
        if (*cp == '\n') {
            if (ejsSetProperty(ejs, (EjsVar*) result, lineno++, 
                    (EjsVar*) ejsCreateStringWithLength(ejs, start, (int) (cp - start))) < 0) {
                break;
            }
            start = cp + 1;
        } else if (*cp == '\r') {
            start = cp + 1;
        }
    }
    if (cp > start) {
        ejsSetProperty(ejs, (EjsVar*) result, lineno++, (EjsVar*) ejsCreateStringWithLength(ejs, start, (int) (cp - start)));
    }

    mprFree(file);
    mprFree(data);

    return (EjsVar*) result;
}


/**
 *  Read the file contents as a string
 *
 *  static function readString(path: String): String
 */
static EjsVar *readFileAsString(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprFile     *file;
    MprBuf      *data;
    EjsVar      *result;
    cchar       *path;
    char        buffer[MPR_BUFSIZE];
    int         bytes, rc;

    mprAssert(argc == 1 && ejsIsString(argv[0]));
    path = ejsGetString(argv[0]);

    file = mprOpen(ejs, path, O_RDONLY | O_BINARY, 0);
    if (file == 0) {
        ejsThrowIOError(ejs, "Can't open %s", path);
        return 0;
    }

    /*
     *  TODO - need to be smarter about running out of memory here if the file is very large.
     */
    data = mprCreateBuf(ejs, 0, (int) mprGetFileSize(file) + 1);
    if (data == 0) {
        ejsThrowMemoryError(ejs);
        return 0;
    }

    rc = 0;
    while ((bytes = mprRead(file, buffer, MPR_BUFSIZE)) > 0) {
        if (mprPutBlockToBuf(data, buffer, bytes) != bytes) {
            ejsThrowMemoryError(ejs);
            rc = -1;
            break;
        }
    }

    result = (EjsVar*) ejsCreateStringWithLength(ejs, mprGetBufStart(data),  mprGetBufLength(data));

    mprFree(file);
    mprFree(data);
    return result;
}


/*
 *  Get the file contents as an XML object
 *
 *  static function readXML(path: String): XML
 */
static EjsVar *readXML(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return 0;
}
#endif

/*
 *  Return a relative path name for the file.
 *
 *  function get relativePath(): Path
 */
static EjsVar *relativePath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprGetRelPath(fp, fp->path));
}


/*
 *  Remove the file associated with the File object.
 *
 *  function remove(): void
 */
static EjsVar *removePathFile(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    if (mprPathExists(ejs, fp->path, R_OK) && unlink(fp->path) < 0) {
        ejsThrowIOError(ejs, "Cant remove file %s", fp->path);
    }
    return 0;
}


/*
 *  function removeDir(): Void
 */
static EjsVar *removePathDir(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprPath     info;

    if (mprGetPathInfo(ejs, fp->path, &info) == 0 && mprDeletePath(ejs, fp->path) < 0) {
        ejsThrowIOError(ejs, "Cant remove directory %s", fp->path);
        return 0;
    }
    return 0;
}


/*
 *  Rename the file
 *
 *  function rename(to: String): Void
 */
static EjsVar *renamePathFile(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    char    *to;

    mprAssert(argc == 1 && ejsIsString(argv[0]));
    to = ejsGetString(argv[0]);

    unlink(to);
    if (rename(fp->path, to) < 0) {
        ejsThrowIOError(ejs, "Cant rename file %s to %s", fp->path, to);
        return 0;
    }
    return 0;
}


static EjsVar *getRepresentation(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    //  TODO
    return 0;
}


static EjsVar *setRepresentation(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    //  TODO
    return 0;
}


/*
 *  Return true if the paths refer to the same file.
 *
 *  function same(other: Object): Boolean
 */
static EjsVar *isPathSame(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) (mprSamePath(ejs, fp->path, ejsGetString(argv[0])) ? ejs->trueValue : ejs->falseValue);
}


/*
 *  function get separator(): String
 */
static EjsVar *pathSeparator(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprFileSystem   *fs;
    EjsString       *sp;

    fs = mprLookupFileSystem(ejs, fp->path);
    sp = ejsCreateString(ejs, fs->separators);
    return (EjsVar*) sp;
}


/*
 *  Get the size of the file associated with this Path
 *
 *  intrinsic function get size(): Number
 */
static EjsVar *getPathFileSize(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    if (mprGetPathInfo(ejs, fp->path, &fp->info) < 0) {
        return (EjsVar*) ejs->minusOneValue;
    }
    return (EjsVar*) ejsCreateNumber(ejs, (MprNumber) fp->info.size);
}


/*
 *  function toString(): String
 */
static EjsVar *pathToJson(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreateStringAndFree(ejs, mprStrcat(ejs, -1, "\"", fp->path, "\"", NULL));
}

/*
 *  function toString(): String
 */
static EjsVar *pathToString(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreateString(ejs, fp->path);
}


/*
 *  function trimExt(): Path
 */
static EjsVar *trimExt(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    return (EjsVar*) ejsCreatePathAndFree(ejs, mprTrimPathExtension(ejs, fp->path));
}


/*
 *  function truncate(size: Number): Void
 */
static EjsVar *truncatePath(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    int     size;

    size = ejsGetInt(argv[0]);
    if (mprTruncatePath(ejs, fp->path, size) < 0) {
        ejsThrowIOError(ejs, "Cant truncate %s", fp->path);
    }
    return 0;
}


#if UNUSED
/*
 *  Put the file contents
 *
 *  static function write(path: String, permissions: Number, ...args): void
 */
static EjsVar *writeToFile(Ejs *ejs, EjsPath *fp, int argc, EjsVar **argv)
{
    MprFile     *file;
    EjsArray    *args;
    char        *path, *data;
    int         i, bytes, length, permissions;

    mprAssert(argc == 3);

    path = ejsGetString(argv[0]);
    permissions = ejsGetInt(argv[1]);
    args = (EjsArray*) argv[2];

    /*
     *  Create fails if already present
     */
    mprDeletePath(ejs, path);
    file = mprOpen(ejs, path, O_CREAT | O_WRONLY | O_BINARY, permissions);
    if (file == 0) {
        ejsThrowIOError(ejs, "Cant create %s", path);
        return 0;
    }

    for (i = 0; i < args->length; i++) {
        data = ejsGetString(ejsToString(ejs, ejsGetProperty(ejs, (EjsVar*) args, i)));
        length = (int) strlen(data);
        bytes = mprWrite(file, data, length);
        if (bytes != length) {
            ejsThrowIOError(ejs, "Write error to %s", path);
            break;
        }
    }
    mprFree(file);
    return 0;
}
#endif


static char *getPath(Ejs *ejs, EjsVar *vp)
{
    if (ejsIsString(vp)) {
        return ejsGetString(vp);
    } else if (ejsIsPath(vp)) {
        return ((EjsPath*) vp)->path;
    }
    ejsThrowIOError(ejs, "Bad path");
    return NULL;
}

/*********************************** Factory **********************************/

EjsPath *ejsCreatePath(Ejs *ejs, cchar *path)
{
    EjsPath     *fp;
    EjsVar      *arg;

    fp = (EjsPath*) ejsCreateVar(ejs, ejsGetType(ejs, ES_ejs_io_Path), 0);
    if (fp == 0) {
        return 0;
    }
    arg = (EjsVar*) ejsCreateString(ejs, path);
    pathConstructor(ejs, fp, 1, (EjsVar**) &arg);
    return fp;
}


EjsPath *ejsCreatePathAndFree(Ejs *ejs, char *value)
{
    EjsPath     *path;

    path = ejsCreatePath(ejs, value);
    mprFree(value);
    return path;
}


void ejsCreatePathType(Ejs *ejs)
{
    EjsName     qname;

    ejsCreateCoreType(ejs, ejsName(&qname, "ejs.io", "Path"), ejs->objectType, sizeof(EjsPath), ES_ejs_io_Path,
        ES_ejs_io_Path_NUM_CLASS_PROP, ES_ejs_io_Path_NUM_INSTANCE_PROP, 
        EJS_ATTR_NATIVE | EJS_ATTR_OBJECT | EJS_ATTR_HAS_CONSTRUCTOR | EJS_ATTR_OBJECT_HELPERS);
}


void ejsConfigurePathType(Ejs *ejs)
{
    EjsType     *type;

    if ((type = ejsGetType(ejs, ES_ejs_io_Path)) == 0) {
        return;
    }

    ejsBindMethod(ejs, type, ES_ejs_io_Path_Path, (EjsNativeFunction) pathConstructor);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_absolute, (EjsNativeFunction) absolutePath);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_accessed, (EjsNativeFunction) getAccessedDate);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_basename, (EjsNativeFunction) getPathBasename);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_copy, (EjsNativeFunction) copyPath);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_created, (EjsNativeFunction) getCreatedDate);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_dirname, (EjsNativeFunction) getPathDirname);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_exists, (EjsNativeFunction) getPathExists);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_extension, (EjsNativeFunction) getPathExtension);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_files, (EjsNativeFunction) getPathFiles);
    ejsBindMethod(ejs, type, ES_Object_get, (EjsNativeFunction) getPathIterator);
    ejsBindMethod(ejs, type, ES_Object_getValues, (EjsNativeFunction) getPathValues);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_hasDrive, (EjsNativeFunction) pathHasDrive);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_isAbsolute, (EjsNativeFunction) isPathAbsolute);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_isDir, (EjsNativeFunction) isPathDir);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_isLink, (EjsNativeFunction) isPathLink);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_isRegular, (EjsNativeFunction) isPathRegular);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_isRelative, (EjsNativeFunction) isPathRelative);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_isSymbolic, (EjsNativeFunction) isPathSymbolic);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_join, (EjsNativeFunction) joinPath);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_joinExt, (EjsNativeFunction) joinPathExt);
    ejsBindMethod(ejs, type, ES_Object_length, (EjsNativeFunction) pathLength);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_makeDir, (EjsNativeFunction) makePathDir);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_makeLink, (EjsNativeFunction) makePathLink);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_makeTemp, (EjsNativeFunction) makePathTemp);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_modified, (EjsNativeFunction) getModifiedDate);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_name, (EjsNativeFunction) getPathName);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_natural, (EjsNativeFunction) getNaturalPath);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_normalize, (EjsNativeFunction) normalizePath);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_parent, (EjsNativeFunction) getPathParent);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_portable, (EjsNativeFunction) getPortablePath);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_relative, (EjsNativeFunction) relativePath);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_remove, (EjsNativeFunction) removePathFile);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_removeDir, (EjsNativeFunction) removePathDir);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_rename, (EjsNativeFunction) renamePathFile);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_representation, (EjsNativeFunction) getRepresentation);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_same, (EjsNativeFunction) isPathSame);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_set_representation, (EjsNativeFunction) setRepresentation);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_separator, (EjsNativeFunction) pathSeparator);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_size, (EjsNativeFunction) getPathFileSize);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_trimExt, (EjsNativeFunction) trimExt);
    ejsBindMethod(ejs, type, ES_ejs_io_Path_truncate, (EjsNativeFunction) truncatePath);

    ejsBindMethod(ejs, type, ES_Object_toJSON, (EjsNativeFunction) pathToJson);
    ejsBindMethod(ejs, type, ES_Object_toString, (EjsNativeFunction) pathToString);
}


#endif /* ES_Path */
/*
 *  @copy   default
 *
 *  Copyright (c) Embedthis Software LLC, 2003-2009. All Rights Reserved.
 *  Copyright (c) Michael O'Brien, 1993-2009. All Rights Reserved.
 *
 *  This software is distributed under commercial and open source licenses.
 *  You may use the GPL open source license described below or you may acquire
 *  a commercial license from Embedthis Software. You agree to be fully bound
 *  by the terms of either license. Consult the LICENSE.TXT distributed with
 *  this software for full details.
 *
 *  This software is open source; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version. See the GNU General Public License for more
 *  details at: http://www.embedthis.com/downloads/gplLicense.html
 *
 *  This program is distributed WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  This GPL license does NOT permit incorporating this software into
 *  proprietary programs. If you are unable to comply with the GPL, you must
 *  acquire a commercial license to use this software. Commercial licenses
 *  for this software and support services are available from Embedthis
 *  Software at http://www.embedthis.com
 *
 *  @end
 */
